/*
 * pi3's Linux kernel Runtime Guard
 *
 * Component:
 *  - Intercept compat_delete_module syscall
 *
 * Notes:
 *  - Enforce Exploit Detection validation
 *    COMPAT version of this syscall is only available in the kernels
 *    with a new syscall implementation (4.17+)
 *
 * Caveats:
 *  - None
 *
 * Timeline:
 *  - Created: 13.VIII.2018
 *
 * Author:
 *  - Adam 'pi3' Zabrocki (http://pi3.com.pl)
 *
 */

#ifdef CONFIG_X86

#ifdef CONFIG_COMPAT

#include "../../../../../p_lkrg_main.h"

#ifdef P_SYSCALL_LAYOUT_4_17


char p_compat_sys_delete_module_kretprobe_state = 0x0;

static struct kretprobe p_compat_sys_delete_module_kretprobe = {
    .kp.symbol_name = P_GET_NEW_COMPAT_SYSCALL_NAME(delete_module),
    .handler = p_compat_sys_delete_module_ret,
    .entry_handler = p_compat_sys_delete_module_entry,
    .data_size = sizeof(struct p_compat_sys_delete_module_data),
    /* Probe up to 40 instances concurrently. */
    .maxactive = 40,
};


int p_compat_sys_delete_module_entry(struct kretprobe_instance *p_ri, struct pt_regs *p_regs) {

   struct p_ed_process *p_tmp;
   unsigned long p_flags;

   p_debug_kprobe_log(
          "Entering function <p_compat_sys_delete_module_entry>\n");

   spin_lock_irqsave(&p_rb_ed_pids_lock, p_flags);
   if ( (p_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(current))) != NULL) {
      // This process is on the ED list - validate 'off' flag
      p_ed_is_off_off_wrap(p_tmp);
   }
   spin_unlock_irqrestore(&p_rb_ed_pids_lock, p_flags);

   p_ed_enforce_validation();
   if (!capable(CAP_SYS_MODULE)) {
      p_print_log(P_LKRG_CRIT,
             "<Exploit Detection> Process[%s | %d] is trying to remove kernel module "
             "but does NOT have appropriate permissions! Killing...\n",
             current->comm, task_pid_nr(current));

      p_regs_set_arg1(p_regs, -1);
      p_regs_set_arg2(p_regs, -1);

      // kill this process!
      p_ed_kill_task_by_task(current);
   }

   p_debug_kprobe_log(
          "Leaving function <p_compat_sys_delete_module_entry>\n");

   return 0x0;
}


int p_compat_sys_delete_module_ret(struct kretprobe_instance *ri, struct pt_regs *p_regs) {

   p_debug_kprobe_log(
          "Entering function <p_compat_sys_delete_module_ret>\n");

   p_ed_enforce_validation();
   if (!capable(CAP_SYS_MODULE)) {

      p_print_log(P_LKRG_CRIT,
             "<Exploit Detection> Process[%s | %d] removed kernel module but does NOT have "
             "appropriate permissions! Killing...\n",
             current->comm, task_pid_nr(current));

      // kill this process!
      p_ed_kill_task_by_task(current);
   }

   p_debug_kprobe_log(
          "Leaving function <p_compat_sys_delete_module_ret>\n");

   return 0x0;
}


int p_install_compat_sys_delete_module_hook(void) {

   int p_ret;

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Entering function <p_install_compat_sys_delete_module_hook>\n");

   if ( (p_ret = register_kretprobe(&p_compat_sys_delete_module_kretprobe)) < 0) {
      p_print_log(P_LKRG_ERR, "[kretprobe] register_kretprobe() for <%s> failed! [err=%d]\n",
                  p_compat_sys_delete_module_kretprobe.kp.symbol_name,
                  p_ret);
      goto p_install_compat_sys_delete_module_hook_out;
   }
   p_print_log(P_LKRG_INFO, "Planted [kretprobe] <%s> at: %p\n",
               p_compat_sys_delete_module_kretprobe.kp.symbol_name,
               p_compat_sys_delete_module_kretprobe.kp.addr);
   p_compat_sys_delete_module_kretprobe_state = 0x1;

//   p_ret = 0x0; <- should be 0x0 anyway...

p_install_compat_sys_delete_module_hook_out:

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Leaving function <p_install_compat_sys_delete_module_hook> (p_ret => %d)\n",p_ret);

   return p_ret;
}


void p_uninstall_compat_sys_delete_module_hook(void) {

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Entering function <p_uninstall_compat_sys_delete_module_hook>\n");

   if (!p_compat_sys_delete_module_kretprobe_state) {
      p_print_log(P_LKRG_INFO, "[kretprobe] <%s> at 0x%p is NOT installed\n",
                  p_compat_sys_delete_module_kretprobe.kp.symbol_name,
                  p_compat_sys_delete_module_kretprobe.kp.addr);
   } else {
      unregister_kretprobe(&p_compat_sys_delete_module_kretprobe);
      p_print_log(P_LKRG_INFO, "Removing [kretprobe] <%s> at 0x%p nmissed[%d]\n",
                  p_compat_sys_delete_module_kretprobe.kp.symbol_name,
                  p_compat_sys_delete_module_kretprobe.kp.addr,
                  p_compat_sys_delete_module_kretprobe.nmissed);
      p_compat_sys_delete_module_kretprobe_state = 0x0;
   }

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Leaving function <p_uninstall_compat_sys_delete_module_hook>\n");
}

#endif

#endif

#endif
