/*
 * pi3's Linux kernel Runtime Guard
 *
 * Component:
 *  - Intercept 'generic_permission' function
 *
 * Notes:
 *  - Enforce Exploit Detection validation
 *
 * Caveats:
 *  - Originally, this file was placing hooks on 'may_open' function.
 *    Unfortunately, GCC (8+) might enable ISRA optimization when -Ox
 *    switch was used. During kernel compilation it is usually enabled,
 *    and as a side effect we have ISRA optimization as well
 *    (undesired for LKRG). ISRA performs interprocedural scalar
 *    replacement of aggregates, removal of unused parameters and
 *    replacement of parameters passed by reference by parameters passed
 *    by value. Since it's a very invasive modification ISRA changes
 *    symbol name of the functions which was modified.
 *    Alexander (Solar Designer) pointed out that in fact we could hook
 *    inode_permission() instead. This function is exported and
 *    additionally, we'll improve our coverage since it is called on more
 *    cases than 'may_open', including things such as permission checks
 *    on creating/removing of directories, (un)linking of files, and
 *    searching for files in directories with restricted permissions.
 *    LKRG hooks 'generic_permission' since this function is also exported
 *    and is called by inode_permission() after various checks have been
 *    made. It is also called in a few other specialized cases.
 *
 * Timeline:
 *  - Replace 'may_open' with 'generic_permission': 17.IX.2018
 *  - Created: 04.X.2017
 *
 * Author:
 *  - Adam 'pi3' Zabrocki (http://pi3.com.pl)
 *
 */

#include "../../../../p_lkrg_main.h"


char p_generic_permission_kretprobe_state = 0x0;

static struct kretprobe p_generic_permission_kretprobe = {
    .kp.symbol_name = "generic_permission",
    .handler = p_generic_permission_ret,
    .entry_handler = p_generic_permission_entry,
    .data_size = sizeof(struct p_generic_permission_data),
    /* Probe up to 40 instances concurrently. */
    .maxactive = 40,
};

/*
 * x86-64 syscall ABI:
 *  *rax - syscall_number
 *    rdi - 1st argument
 *    rsi - 2nd argument
 *    rdx - 3rd argument
 *    rcx - 4th argument
 *
 *    r8  - 5th one
 *    r9  - 6th one
 */

int p_generic_permission_entry(struct kretprobe_instance *p_ri, struct pt_regs *p_regs) {

   struct p_ed_process *p_tmp = NULL;
   unsigned long p_flags;

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Entering function <p_generic_permission_entry>\n");

//   p_ed_enforce_validation();

   if (p_is_ed_task(current)) {
      spin_lock_irqsave(&p_rb_ed_pids_lock, p_flags);
      p_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(current));
      if (p_ed_pcfi_validate_sp(current, p_tmp, p_regs_get_sp(p_regs))) {
         p_print_log(P_LKRG_CRIT,
                "<Exploit Detection> Stack pointer corruption (ROP?) - pCFI violation: process[%s | %d] !!!\n",
                current->comm,task_pid_nr(current));
         // kill this process!
         p_ed_kill_task_by_task(current);
      }
      p_validate_task_f(current);
      spin_unlock_irqrestore(&p_rb_ed_pids_lock, p_flags);
   }

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Leaving function <p_generic_permission_entry>\n");

   return 0x0;
}


int p_generic_permission_ret(struct kretprobe_instance *p_ri, struct pt_regs *p_regs) {

//   struct p_ed_process *p_tmp = NULL;
//   unsigned long p_flags;

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Entering function <p_generic_permission_ret>\n");

//   p_ed_enforce_validation();
/*
   spin_lock_irqsave(&p_rb_ed_pids_lock, p_flags);
   p_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(current));
   if (p_ed_pcfi_validate_sp(current, p_regs, p_tmp)) {
      p_print_log(P_LKRG_CRIT,
             "<Exploit Detection> Stack pointer corruption (ROP?) - pCFI violation: process[%s | %d] !!!\n",
             current->comm,task_pid_nr(current));
      p_ed_kill_task_by_task(current);
   }
   spin_unlock_irqrestore(&p_rb_ed_pids_lock, p_flags);
*/
// STRONG_DEBUG
   p_debug_kprobe_log(
          "Leaving function <p_generic_permission_ret>\n");

   return 0x0;
}


int p_install_generic_permission_hook(void) {

   int p_ret;

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Entering function <p_install_generic_permission_hook>\n");

   if ( (p_ret = register_kretprobe(&p_generic_permission_kretprobe)) < 0) {
      p_print_log(P_LKRG_ERR, "[kretprobe] register_kretprobe() for <%s> failed! [err=%d]\n",
                  p_generic_permission_kretprobe.kp.symbol_name,
                  p_ret);
      goto p_install_generic_permission_hook_out;
   }
   p_print_log(P_LKRG_INFO, "Planted [kretprobe] <%s> at: %p\n",
               p_generic_permission_kretprobe.kp.symbol_name,
               p_generic_permission_kretprobe.kp.addr);
   p_generic_permission_kretprobe_state = 0x1;

//   p_ret = 0x0; <- should be 0x0 anyway...

p_install_generic_permission_hook_out:

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Leaving function <p_install_generic_permission_hook> (p_ret => %d)\n",p_ret);

   return p_ret;
}


void p_uninstall_generic_permission_hook(void) {

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Entering function <p_uninstall_generic_permission_hook>\n");

   if (!p_generic_permission_kretprobe_state) {
      p_print_log(P_LKRG_INFO, "[kretprobe] <%s> at 0x%p is NOT installed\n",
                  p_generic_permission_kretprobe.kp.symbol_name,
                  p_generic_permission_kretprobe.kp.addr);
   } else {
      unregister_kretprobe(&p_generic_permission_kretprobe);
      p_print_log(P_LKRG_INFO, "Removing [kretprobe] <%s> at 0x%p nmissed[%d]\n",
                  p_generic_permission_kretprobe.kp.symbol_name,
                  p_generic_permission_kretprobe.kp.addr,
                  p_generic_permission_kretprobe.nmissed);
      p_generic_permission_kretprobe_state = 0x0;
   }

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Leaving function <p_uninstall_generic_permission_hook>\n");
}
