/*
 * pi3's Linux kernel Runtime Guard
 *
 * Component:
 *  - Intercept SECCOMP policy update
 *
 * Notes:
 *  - Process SECCOMP Exploit Detection validation
 *
 * Caveats:
 *  - None
 *
 * Timeline:
 *  - Created: 18.XI.2017
 *
 * Author:
 *  - Adam 'pi3' Zabrocki (http://pi3.com.pl)
 *
 */

#include "../../../../p_lkrg_main.h"


char p_seccomp_kretprobe_state = 0x0;

static struct kretprobe p_seccomp_kretprobe = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
    .kp.symbol_name = "do_seccomp",
#else
    .kp.symbol_name = "prctl_set_seccomp",
#endif
    .handler = p_seccomp_ret,
    .entry_handler = p_seccomp_entry,
    .data_size = sizeof(struct p_seccomp_data),
    /* Probe up to 40 instances concurrently. */
    .maxactive = 40,
};

/*
 * x86-64 syscall ABI:
 *  *rax - syscall_number
 *    rdi - 1st argument
 *    rsi - 2nd argument
 *    rdx - 3rd argument
 *    rcx - 4th argument
 *
 *    r8  - 5th one
 *    r9  - 6th one
 */

int p_seccomp_entry(struct kretprobe_instance *p_ri, struct pt_regs *p_regs) {

   struct p_ed_process *p_tmp;
   unsigned long p_flags;

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Entering function <p_seccomp_entry>\n");

//   p_ed_enforce_validation();

   spin_lock_irqsave(&p_rb_ed_pids_lock, p_flags);
   if ( (p_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(current))) != NULL) {
      // This process is on the ED list - set temporary 'disable' flag!
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
      if (p_regs_get_arg2(p_regs) & SECCOMP_FILTER_FLAG_TSYNC) { // SECCOMP_FILTER_FLAG_TSYNC
         struct task_struct *p_father = current;
         struct task_struct *p_threads = p_father;
         struct p_ed_process *p_child_tmp;

         p_tmp->p_ed_task.p_sec.flag_sync_thread = 0x1;

         rcu_read_lock();
         // Available since 3.14.0
         for_each_thread(p_father, p_threads) {
            get_task_struct(p_threads);
            p_child_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(p_threads));
            if (p_child_tmp) {
               p_set_ed_process_off(p_child_tmp);
            }
            put_task_struct(p_threads);
         }
         rcu_read_unlock();
      } else {
#endif
         p_set_ed_process_off(p_tmp);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
      }
#endif
   }
   spin_unlock_irqrestore(&p_rb_ed_pids_lock, p_flags);

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Leaving function <p_seccomp_entry>\n");

   return 0x0;
}


int p_seccomp_ret(struct kretprobe_instance *p_ri, struct pt_regs *p_regs) {

   struct p_ed_process *p_tmp;
   char p_update = ((long)p_regs_get_ret(p_regs) >= 0) ? 1 : 0;
   unsigned long p_flags;

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Entering function <p_seccomp_ret>\n");
   // Update process
   spin_lock_irqsave(&p_rb_ed_pids_lock, p_flags);
   if ( (p_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(current))) != NULL) {
      // This process is on the ED list - update information!
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
      if (p_tmp->p_ed_task.p_sec.flag_sync_thread) { // SECCOMP_FILTER_FLAG_TSYNC
         struct task_struct *p_father = current;
         struct task_struct *p_threads = p_father;
         struct p_ed_process *p_child_tmp;

         rcu_read_lock();
         // Available since 3.14.0
         for_each_thread(p_father, p_threads) {
            get_task_struct(p_threads);
            p_child_tmp = p_rb_find_ed_pid(&p_global_ed_pids_root, task_pid_nr(p_threads));
            if (p_child_tmp) {
               if (p_update) {
                  p_print_log(P_LKRG_INFO, "Updating ED pid[%d]\n",p_threads->pid);
                  p_update_ed_process(p_child_tmp, p_threads);
               }
               if (p_threads != p_father)
                  p_set_ed_process_on(p_child_tmp);
            }
            put_task_struct(p_threads);
         }
         rcu_read_unlock();
      } else {
#endif
         if (p_update) {
            p_print_log(P_LKRG_INFO, "Updating ED pid[%d]\n",current->pid);
            p_update_ed_process(p_tmp, current);
         }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
      }
#endif
      p_set_ed_process_on(p_tmp);
      // In case of seccomp failing with SECCOMP_FILTER_FLAG_TSYNC flag
      p_tmp->p_ed_task.p_sec.flag_sync_thread = 0x0;
   }
   spin_unlock_irqrestore(&p_rb_ed_pids_lock, p_flags);

//   p_ed_enforce_validation();

// STRONG_DEBUG
   p_debug_kprobe_log(
          "Leaving function <p_seccomp_ret>\n");

   return 0x0;
}


int p_install_seccomp_hook(void) {

   int p_ret;

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Entering function <p_install_seccomp_hook>\n");

   if ( (p_ret = register_kretprobe(&p_seccomp_kretprobe)) < 0) {
      p_print_log(P_LKRG_ERR, "[kretprobe] register_kretprobe() for <%s> failed! [err=%d]\n",
                  p_seccomp_kretprobe.kp.symbol_name,
                  p_ret);
      goto p_install_seccomp_hook_out;
   }
   p_print_log(P_LKRG_INFO, "Planted [kretprobe] <%s> at: %p\n",
               p_seccomp_kretprobe.kp.symbol_name,
               p_seccomp_kretprobe.kp.addr);
   p_seccomp_kretprobe_state = 0x1;

//   p_ret = 0x0; <- should be 0x0 anyway...

p_install_seccomp_hook_out:

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Leaving function <p_install_seccomp_hook> (p_ret => %d)\n",p_ret);

   return p_ret;
}


void p_uninstall_seccomp_hook(void) {

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Entering function <p_uninstall_seccomp_hook>\n");

   if (!p_seccomp_kretprobe_state) {
      p_print_log(P_LKRG_INFO, "[kretprobe] <%s> at 0x%p is NOT installed\n",
                  p_seccomp_kretprobe.kp.symbol_name,
                  p_seccomp_kretprobe.kp.addr);
   } else {
      unregister_kretprobe(&p_seccomp_kretprobe);
      p_print_log(P_LKRG_INFO, "Removing [kretprobe] <%s> at 0x%p nmissed[%d]\n",
                  p_seccomp_kretprobe.kp.symbol_name,
                  p_seccomp_kretprobe.kp.addr,
                  p_seccomp_kretprobe.nmissed);
      p_seccomp_kretprobe_state = 0x0;
   }

// STRONG_DEBUG
   p_debug_log(P_LKRG_STRONG_DBG,
          "Leaving function <p_uninstall_seccomp_hook>\n");
}
